/****************************************************************
VELOCITY.PRG

This is a GAUSS program, written by Bruce E. Hansen.
It replicates Table 4 and Figures 1and 2 from my paper
"The Grid Bootstrap and the Autoregressive Model".

If you have questions, please contact me at

Bruce E. Hansen
Department of Economics
Social Science Building
University of Wisconsin
Madison, WI 53706-1393
bhansen@ssc.wisc.edu
*/
/*
g=200; 	@ 200@
b=1999;    @1999@
bq=1999;   @ 1999@
c=.9;           @ confidence level @
width=6;
if c==.9;  aa=1.645; endif;
cc=((1-c)/2)|((c+1)/2);
rc=rev(cc);
load stock_t;

load np[]=nplosser.dat;
y  = np[1215:1334];
n=rows(y);

xs=1869;
x=seqa(xs,1,n);
graphset;
_pdate="";
title("Velocity: 1869-1988");
xy(x,y);

/*****************************************************************/
ff1 = "#*.*lG"~16~8;

output file=velocity.out reset;
"Nelson-Plosser Velocity";
"Sample:   1869-1988";
"Gridpoints                 " g;
"Standard Bootstraps " b;
"Grid Bootstraps        " bq;
"Confidence Level      " c;
"";
determ=ones(n-1,1)~seqa(1,1,n-1);
d0=(1~0)|determ;
yt=y-d0*(y/d0);
y0=yt[1];
xname="Asymptotic"|"Stock"|"Efron"|"Perct-t"|
"Kilian"|"Grid-a"|"Grid-t";
nt=rows(xname);

{theta,se,e}=b_est(y);
ahat=theta[3];sehat=se[3];
pr=printfm(("Variable"~"Estimate"~"S.E."),0~0~0,ff1);"";
pr=printfm(("Intercpt"|"Trend"|"Y(-1)")~theta~se,0~1~1,ff1);"";

ends=zeros(nt,2);
@ Asymptotic methods @
ui=ahat+((-1)|1)*sehat*aa;
ends[1,.]=ui';
ui=unitci((ahat-1)/sehat,c,stock_t);   @ stock @
ends[2,.]=(1+ui/(n-1))';

@ Standard Bootstraps @
boota=zeros(b,1);
boott=zeros(b,1);
j=1; do while j<=b;
  {thetab,seb,eb}=b_est(ar1(y0,ahat,e));
   boota[j]=thetab[3];
   boott[j]=(thetab[3]-ahat)/seb[3];
j=j+1;endo;
@ efron @
ef=quantile(boota,cc);
ends[3,.]=ef;
@ percentile t @
qt=quantile(boott,rc);
pt=ahat-qt*sehat;
ends[4,.]=pt;

@ Kilian bootstrap @
if ahat>=1;
  bc=ahat;bias=0;
else;
  bias=meanc(boota)-ahat; bc=ahat-bias;
  if bc>=1;bc=1;bias=ahat-bc;endif;
endif;
j=1; do while j<=b;
  {thetab,seb,eb}=b_est(ar1(y0,bc,e));
   boota[j]=thetab[3]-bias;
j=j+1;endo;
ends[5,.]=quantile(boota,cc);

@ Grid Bootstrap @
bootq=zeros(bq,g);
bootse=zeros(bq,g);
a0=seqa(.84,.22/g,g);
ak=1; do while ak<=g;
  alphak=a0[ak];
  j=1; do while j<=bq;
    {thetab,seb,eb}=b_est(ar1(y0,alphak,e));
    bootq[j,ak]=thetab[3];    bootse[j,ak]=seb[3];
  j=j+1;endo;
ak=ak+1;endo;
ba=bootq-a0';
ts=ba./bootse;
grid_p=kq_reg(ba,a0,cc);
grid_t=kq_reg(ts,a0,cc);
ah0=ahat-a0;
ts=ah0/sehat;
ends[6,.]=g_end(a0,ah0,grid_p);
ends[7,.]=g_end(a0,ts,grid_t);
"";
"Confidence Intervals for AR(1) Parameter";
pr=printfm("Method"~"Left End"~"Right End",0~0~0,ff1);"";
pr=printfm(xname~ends,0~1~1,ff1);"";
"________________________________________________________________";
"";
output off;

/*
@ Plot 1 - Efron Construction @
z=zeros(g,1);
eu=ef[2]-ahat;el=ef[1]-ahat;
l=ends[6,1];u=ends[6,2];
ql=ahat-l;qu=ahat-u;
hpu=ahat+eu;hpl=ahat+el;
graphset;
_pdate="";
fonts("complex simgrma");
_plwidth=2;
let _pltype = 6 6 3 1;
title("Velocity: 1869-1988");
ylabel("Bootstrap Quantile");
xlabel("90% Percentile Confidence Interval");
xtics(.84,1,.02,2);
ytics(-.15,.05,.05,5);
_psym=
(ahat~0~8~3~15~1~0)
|(ahat~eu~5~3~15~1~0)
|(ahat~el~5~3~15~1~0)
|(hpu~eu~2~3~15~1~0)
|(hpl~el~2~3~15~1~0)
;
_pline=
(1~4~hpl~el~ahat~el~1~15~5)
|(1~4~ahat~el~ahat~eu~0~15~5)
|(1~4~ahat~eu~hpu~eu~0~15~5)
;
_parrow=
(hpu~eu~hpu~0~4~.2~11~15~1~4~5)
|(hpl~el~hpl~0~4~.2~11~15~1~4~5)
;
xy(a0,z~(a0-ahat)~grid_p[.,2 1]);
*/
*/

@ Plot 2 - Grid-alpha Construction (Figure 2) @
graphset;
z=zeros(g,1);
eu=ef[2]-ahat;el=ef[1]-ahat;
l=ends[6,1];u=ends[6,2];
ql=ahat-l;qu=ahat-u;
_pdate="";
_pcolor=15;
fonts("complex simgrma");
_plwidth=2;
let _pltype = 6 6 3 1;
title("Figure 2\LVelocity: 1869-1988");
ylabel("Bootstrap Quantile");
xtics(.9,1.06,.02,2);
ytics(-.20,.05,.05,5);
xlabel("90% Grid-alpha Confidence Interval");
_psym=
(ahat~0~8~3~15~1~0)
|(l~ql~1~3~15~1~0)
|(u~qu~1~3~15~1~0)
;
_parrow=
(l~ql~l~0~4~.2~01~15~1~6~2)
|(u~qu~u~0~4~.2~01~15~1~6~2);
xy(a0,ah0~z~grid_p[.,2 1]);
_protate=1;
graphprt("-c=1 -cf=fig_2.eps");
xy(a0,ah0~z~grid_p[.,2 1]);


@ Plot 3 - Percentile-t and Grid-t Construction (Figure 1) @
graphset;
_pdate="";
fonts("complex simgrma");
_plwidth=2;
let _pltype = 6 6 3 1;
_pcolor=15;
ylabel("Bootstrap Quantile");
let _pltype = 6 6 3 1;
l=ends[7,1];u=ends[7,2];
eu=qt[1];el=qt[2];
ql=(ahat-l)/sehat;qu=(ahat-u)/sehat;
hl=ahat-sehat*el;
hu=ahat-sehat*eu;
_psym=
(ahat~0~8~3~15~1~0)
|(ahat~eu~5~3~15~1~0)
|(ahat~el~5~3~15~1~0)
|(hl~el~2~3~15~1~0)
|(hu~eu~2~3~15~1~0)
|(l~ql~1~3~15~1~0)
|(u~qu~1~3~15~1~0);
_pline=
(1~4~hl~el~ahat~el~1~15~5)
|(1~4~ahat~el~ahat~eu~0~15~5)
|(1~4~ahat~eu~hu~eu~0~15~5);
_parrow=
(l~ql~l~0~4~.2~01~15~1~6~2)
|(u~qu~u~0~4~.2~01~15~1~6~2)
|(hu~eu~hu~0~4~.2~11~15~1~4~5)
|(hl~el~hl~0~4~.2~11~15~1~4~5);
xtics(.9,1.06,.02,2);
ytics(-4,2,1,2);
xlabel("90% Percentile-t and Grid-t Confidence Intervals");
title("Figure 1\LVelocity: 1869-1988");
xy(a0,ts~z~grid_t[.,2 1]);
_protate=1;
graphprt("-c=1 -cf=fig_1.eps");
xy(a0,ts~z~grid_t[.,2 1]);

/****************************************************************/

/*** Procedures ***/

proc (3) = b_est(yy);
local x,y,xx,theta,e,se;
  x=determ~yy[1:n-1];
  y=yy[2:n];
  xx=invpd(moment(x,0));
  theta=xx*(x'y);
  e=y-x*theta;
  se=sqrt(diag(xx*(e'e)/(n-4)));
retp(theta,se,e);
endp;


proc ar1(y0,a,e);
local y,nn,y00;
  nn=rows(e);
  if abs(a)<1; y00=y0;else;y00=0;endif;
  y=recserar(e[ceil(nn*rndu(n,1))],y00,a);
retp(y);
endp;


proc (2) = cross(a,x,y);
local ms,rn,r,d,s,li,al,bl,au,bu,g1,g2;
  ms=0;  ms=miss(ms,ms);
  rn=rows(a);
  r=(x.<y);
  d=0|(r[2:rn]-r[1:rn-1]);
  s=seqa(1,1,rn);
  li=selif(s,(d .== 1));
  if maxc(li) .ne ms;
    al=a[li-1];au=a[li];
    bl=x[li-1]-y[li-1];   bu=y[li]-x[li];
    g1=(bu.*al+bl.*au)./(bl+bu);
  else;
    g1=ms;
  endif;
  li=selif(s,(d .== -1));
  if maxc(li) .ne ms;
    al=a[li-1];au=a[li];
    bl=y[li-1]-x[li-1];   bu=x[li]-y[li];
    g2=(bu.*al+bl.*au)./(bl+bu);
  else;
    g2=ms;
  endif;
retp(g1,g2);
endp;


proc g_end(a0,ta,grid);
local l1,l2,u1,u2,s,e1,e2,length,u,l;
  {l1,l2}=cross(a0,ta,grid[.,1]);
  {u1,u2}=cross(a0,ta,grid[.,2]);
  s=packr(sortc(l1|l2|u1|u2,1));
  l=minc(s);
  u=maxc(s);
  if l==minc(u2); l=minc(a0); endif;
  if l==minc(l1); l=minc(a0); endif;
  if u==maxc(l2); u=maxc(a0); endif;
  if u==maxc(u1); u=maxc(a0); endif;
retp(l~u);
endp;


proc quantile(x,c);
local b,r,xx,j,bc,i,q;
  b=rows(x); r=cols(x);xx=x;
  j=1; do while j<=r;
    xx[.,j]=sortc(x[.,j],1);
  j=j+1;endo;
  bc=b*c;
  i=floor(bc);
  i=i+(i .== 0);
  q=xx[i,.].*(i .== bc) + ((xx[i,.]+xx[i+1,.])/2).*(i .ne bc);
retp(q');
endp;


proc (3) = c_out(cc,alpha);
retp((cc[1] .> alpha),(cc[2].<alpha),(cc[2]-cc[1]));
endp;


proc epi(q,a,h,leave);
local u,w;
  u=(a-a')/h;
  w=(1-u.^2).*(abs(u) .<= 1);
  if leave==1;
    w=diagrv(w,zeros(rows(a),1));
  endif;
  w=w./sumc(w');
retp(w*q);
endp;


proc kq_reg(y,x,c);
local hn,hh,cn,check,q,hi,qh,h,ci;
  hn=5;
  cn=rows(c);
  hh=seqa(1,1,hn)*(x[2]-x[1]);
  check=zeros(hn,cn);
  q=quantile(y,c);
  hi=1; do while hi<=hn;
    h=hh[hi];
    if x[2]-x[1]<h;
      qh=epi(q,x,h,1);
      check[hi,.]=sumc((q-qh).^2)';
     else;check[hi,.]=sumc(q.^2)';
     endif;
  hi=hi+1;endo;
  hi=minindc(check);
  h=hh[hi];
  qh=zeros(g,cn);
  ci=1; do while ci<=cn;
    qh[.,ci]=epi(q[.,ci],x,h[ci],0);
  ci=ci+1;endo;
retp(qh);
endp;


@ Stock's proc for computing local-to unity confidence intervals @
proc(1) = unitci(tdf,clevel,cv);
 local alpha0,sel0,sel1,i0,i1,cc,cv0,cv1,d1,d2,dd,jj,j0,j1,wt0,wt1,c0,c1;
 alpha0=0.5*(1-clevel);
 sel0=(fix(1000*cv[1,.]'+.01).==fix(1000*alpha0+.01));
 sel1=(fix(1000*cv[1,.]'+.01).==fix(1000*(1-alpha0)+.01));
 i0=sumc(seqa(1,1,cols(cv)).*sel0);
 i1=sumc(seqa(1,1,cols(cv)).*sel1);
 cc=cv[2:rows(cv),1]; cv0=cv[2:rows(cv),i0]; cv1=cv[2:rows(cv),i1];
 d1=(cv0.<tdf); d2=(cv1.>tdf); dd=d1.*d2;
 if i0==i1;    @ return median -- 0-width CI @
  if sumc(d1)>0 and sumc(d2)>0;
   jj=selif(seqa(1,1,rows(cc)),d1); j0=maxc(jj); j1=j0+1;
  elseif sumc(d1)==0; c0=-99.99; c1=c0; goto done;
  elseif sumc(d2)==0; c0=99.99; c1=c0; goto done;
  endif;
 elseif sumc(dd)>0; @ nonempty CI @
  jj=selif(seqa(1,1,rows(cc)),dd); j0=maxc(jj); j1=minc(jj);
 else; @ empty confidence interval -- outside the range covered by cv @
  if sumc(d1)==0; c0=-99.99; c1=-99.99;
  elseif sumc(d2)==0; c0=99.99; c1=99.99;
  endif;
  goto done;
 endif;
 if j0<rows(cc);
  wt0=(tdf-cv0[j0,1])/(cv0[j0+1,1]-cv0[j0,1]);
   c0=(1-wt0)*cc[j0,1] + wt0*cc[j0+1,1];
 else; wt0=0; c0=99.99; endif;
 if j1>1;
  wt1=(tdf-cv1[j1-1,1])/(cv1[j1,1]-cv1[j1-1,1]);
   c1=(1-wt1)*cc[j1-1,1] + wt1*cc[j1,1];
 else; wt1=0; c1=-99.99; endif;
done:
retp(c1|c0);
endp;

